import pytest
import logging
from unittest.mock import Mock
from webtest import TestApp as WSGIApp
from horseman.response import Response
from horseman.meta import Node, SentryNode
from horseman.http import HTTPError


class TestNode:

    def test_direct_instance(self):
        # Can't instanciate directly a class with abstract methods.
        with pytest.raises(TypeError):
            Node()

    def test_call(self):

        class MyNode(Node):

            def resolve(self, path_info, environ):
                if path_info == "/test":
                    return Response(body=b"Some Content")

        node = WSGIApp(MyNode())
        response = node.get("/", expect_errors=True)
        assert response.status == "404 Not Found"

        response = node.get("/test", expect_errors=True)
        assert response.status == "200 OK"
        assert response.body == b"Some Content"

    def test_error_handling(self):

        class MyNode(Node):

            def resolve(self, path_info, environ):
                raise HTTPError(400, 'This is a dead end')

        node = WSGIApp(MyNode())
        response = node.get("/", expect_errors=True)
        assert response.status == "400 Bad Request"
        assert response.body == b'This is a dead end'

    def test_path_normalization(self):

        class MyNode(Node):

            def resolve(self, path_info, environ):
                return Response(body=path_info)

        node = WSGIApp(MyNode())
        assert node.get("/test").body == b'/test'
        assert node.get("///test//").body == b'/test/'
        assert node.get("///foo//bar////").body == b'/foo/bar/'
        assert node.get("").body == b''


class TestSentryNode:

    def test_direct_instance(self):
        with pytest.raises(TypeError) as exc:
            SentryNode()
        assert str(exc.value) == (
            "Can't instantiate abstract class SentryNode "
            "with abstract methods handle_exception, resolve"
        )

    def test_call(self, caplog):

        class MyNode(SentryNode):

            def handle_exception(self, exc_info, environ):
                logging.warning('An error occured !')

            def resolve(self, path_info, environ):
                raise NotImplementedError('I failed !')

        node = WSGIApp(MyNode())
        with caplog.at_level(logging.WARNING):
            with pytest.raises(NotImplementedError):
                node.get("/", expect_errors=True)

        assert len(caplog.records) == 1
        assert caplog.records[0].message == 'An error occured !'

    def test_close(self, caplog):

        class Response:

            def __iter__(self):
                yield b'I am done'

            def close(self):
                logging.info('I closed')

            def __call__(self, environ, start_response):
                start_response('200 OK', [])
                return self

        class MyNode(SentryNode):

            def handle_exception(self, exc_info, environ):
                pass

            def resolve(self, path_info, environ):
                return Response()

        node = WSGIApp(MyNode())
        with caplog.at_level(logging.INFO):
            node.get("/")

        assert len(caplog.records) == 1
        assert caplog.records[0].message == 'I closed'

    def test_close_exception(self, caplog):

        error = NotImplementedError("I don't know what to do !")

        class Response:

            def __iter__(self):
                yield b'I am done'

            def close(self):
                raise error

            def __call__(self, environ, start_response):
                start_response('200 OK', [])
                return self

        handle = Mock()

        class MyNode(SentryNode):

            def handle_exception(self, exc_info, environ):
                cls, exc, tb = exc_info
                handle(exc)

            def resolve(self, path_info, environ):
                return Response()

        node = WSGIApp(MyNode())
        with pytest.raises(NotImplementedError):
            node.get("/")

        assert handle.call_count == 1
        handle.assert_called_once_with(error)



from horseman.http import ContentType


def test_content_type():

    header = (
        '''Message/Partial; number=2; total=3; '''
        '''id="oc=jpbe0M2Yt4s@thumper.bellcore.com";'''
    )

    ct = ContentType.from_http_header(header)
    assert ct.mimetype == "Message/Partial"
    assert ct.options == {
        'number': '2',
        'total': '3',
        'id': 'oc=jpbe0M2Yt4s@thumper.bellcore.com'
    }



from horseman.meta import APIView, Overhead
from horseman.response import Response


class Request(Overhead):

    data = None

    def __init__(self, environ):
        self.environ = environ

    def extract(self):
        self.data = 'somedata'


def test_apiview():

    class View(APIView):

        def GET(self, overhead):
            return Response(200, b'You got it.')

    environ = {'REQUEST_METHOD': 'GET'}
    request = Request(environ)
    response = View()(request)
    assert response.body == b'You got it.'

    environ = {'REQUEST_METHOD': 'POST'}
    request = Request(environ)
    response = View()(request)
    assert response.status.value == 405
    assert next(iter(response)) == (
        b'Specified method is invalid for this resource'
    )


def test_api_view_multiple_methods():

    class View(APIView):

        def GET(self, overhead):
            return Response(body=b'You got it.')

        def POST(self, overhead):
            return Response(body=b'You posted it.')

    environ = {'REQUEST_METHOD': 'GET'}
    request = Request(environ)
    response = View()(request)
    assert response.body == b'You got it.'

    environ = {'REQUEST_METHOD': 'POST'}
    request = Request(environ)
    response = View()(request)
    assert response.status.value == 200
    assert response.body == b'You posted it.'



import re
import orjson
from http import HTTPStatus
from typing import Optional, Union, Dict, List, NamedTuple, IO, Callable
from horseman.datastructures import FormData
from horseman.parsers.multipart import Multipart
from horseman.http import HTTPError, ContentType, Query
from horseman.types import Charset, MIMEType


MIME_TYPE_REGEX = re.compile(r"^multipart|[-\w.]+/[-\w.\+]+$")


class Data(NamedTuple):
    form: Optional[FormData] = None
    files: Optional[FormData] = None
    json: Optional[Union[Dict, List]] = None  # not too specific


Boundary = str
Parser = Callable[[
    IO, MIMEType, Optional[Union[Charset, Boundary]]
], Data]


class BodyParser:

    __slots__ = ('parsers',)

    def __init__(self):
        self.parsers = {}

    def register(self, mimetype: MIMEType):
        if not MIME_TYPE_REGEX.fullmatch(mimetype):
            raise ValueError(f'{mimetype!r} is not a valid MIME Type.')

        def registration(parser: Parser) -> Parser:
            self.parsers[mimetype.lower()] = parser
            return parser
        return registration

    def get(self, mimetype: MIMEType) -> Optional[Parser]:
        return self.parsers.get(mimetype)

    def __call__(self, body: IO, header: Union[str, ContentType]) -> Data:
        if not isinstance(header, ContentType):
            content_type = ContentType.from_http_header(header)
        else:
            content_type = header
        parser = self.get(content_type.mimetype)
        if parser is None:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST,
                f'Unknown content type: {content_type.mimetype!r}.'
            )
        try:
            return parser(
                body, content_type.mimetype, **content_type.options)
        except ValueError as exc:
            raise HTTPError(HTTPStatus.BAD_REQUEST, str(exc))


parser = BodyParser()


@parser.register('application/json')
def json_parser(body: IO, mimetype: MIMEType,
                charset: Charset = 'utf-8') -> Data:
    data = body.read()
    if not data:
        raise ValueError('The body of the request is empty.')
    try:
        jsondata = orjson.loads(data.decode(charset))
        return Data(json=jsondata)
    except orjson.JSONDecodeError:
        raise ValueError('Unparsable JSON body.')


@parser.register('multipart/form-data')
def multipart_parser(body: IO, mimetype: MIMEType,
                     boundary: Optional[str] = None) -> Data:
    if boundary is None:
        raise ValueError('Missing boundary in Content-Type.')
    content_parser = Multipart(f";boundary={boundary}")
    while chunk := body.read(8192):
        try:
            content_parser.feed_data(chunk)
        except ValueError:
            raise ValueError('Unparsable multipart body.')
    return Data(form=content_parser.form, files=content_parser.files)


@parser.register('application/x-www-form-urlencoded')
def urlencoded_parser(body: IO, mimetype: MIMEType,
                      charset: Charset = 'utf-8') -> Data:
    data = body.read()
    if not data:
        raise ValueError('The body of the request is empty.')
    try:
        form = Query.from_value(data.decode(charset))
    except UnicodeDecodeError:
        raise ValueError(f'Failed to decode using charset {charset!r}.')
    return Data(form=form)



import pytest
from io import BytesIO
from horseman.http import Query
from horseman.parsers import Data, urlencoded_parser


def test_empty_urlencoded():
    body = BytesIO(b'')
    with pytest.raises(ValueError) as exc:
        urlencoded_parser(body, 'application/x-www-form-urlencoded')
    assert str(exc.value) == "The body of the request is empty."


def test_urlencoded():
    body = BytesIO(b'name=MacBeth&thane=Cawdor&thane=Glamis')
    data = urlencoded_parser(body, 'application/x-www-form-urlencoded')
    assert isinstance(data, Data)
    assert data.files is None
    assert data.json is None
    assert isinstance(data.form, Query)
    assert tuple(data.form.pairs()) == (
        ('name', 'MacBeth'),
        ('thane', 'Cawdor'),
        ('thane', 'Glamis')
    )


def test_urlencoded_charset():
    body = BytesIO("name=Älfùr".encode('utf-8'))
    data = urlencoded_parser(body, 'application/x-www-form-urlencoded')
    assert isinstance(data, Data)
    assert data.files is None
    assert data.json is None
    assert isinstance(data.form, Query)
    assert tuple(data.form.pairs()) == (
        ('name', 'Älfùr'),
    )

    body = BytesIO("name=Älfùr".encode('latin-1'))
    with pytest.raises(ValueError) as exc:
        urlencoded_parser(body, 'application/x-www-form-urlencoded')
    assert str(exc.value) == "Failed to decode using charset 'utf-8'."

    body = BytesIO("name=Älfùr".encode('latin-1'))
    data = urlencoded_parser(
        body, 'application/x-www-form-urlencoded', charset='latin-1')
    assert isinstance(data, Data)
    assert data.files is None
    assert data.json is None
    assert isinstance(data.form, Query)
    assert tuple(data.form.pairs()) == (
        ('name', 'Älfùr'),
    )


def test_wrong_urlencoded():
    body = BytesIO(b'foo')
    with pytest.raises(ValueError) as exc:
        urlencoded_parser(body, 'application/x-www-form-urlencoded')
    assert str(exc.value) == "bad query field: 'foo'"



import orjson
from http import HTTPStatus
from pathlib import Path
from multidict import CIMultiDict
from horseman.http import Cookies
from horseman.types import Environ, HTTPCode, StartResponse, WSGICallable
from typing import Deque, Callable, Any, Generator, Iterable, Optional, Tuple


BODYLESS = frozenset((
    HTTPStatus.CONTINUE,
    HTTPStatus.SWITCHING_PROTOCOLS,
    HTTPStatus.PROCESSING,
    HTTPStatus.NO_CONTENT,
    HTTPStatus.NOT_MODIFIED
))

REDIRECT = frozenset((
    HTTPStatus.MULTIPLE_CHOICES,
    HTTPStatus.MOVED_PERMANENTLY,
    HTTPStatus.FOUND,
    HTTPStatus.SEE_OTHER,
    HTTPStatus.NOT_MODIFIED,
    HTTPStatus.USE_PROXY,
    HTTPStatus.TEMPORARY_REDIRECT,
    HTTPStatus.PERMANENT_REDIRECT
))


def file_iterator(path: Path, chunk: int = 4096):
    with path.open('rb') as reader:
        while True:
            data = reader.read(chunk)
            if not data:
                break
            yield data


class Headers(CIMultiDict):

    cookies: Cookies

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cookies = None

    @property
    def cookies(self):
        if self._cookies is None:
            self._cookies = Cookies()
        return self._cookies

    def items(self) -> Generator[Tuple[str, str], None, None]:
        yield from super().items()
        if self._cookies:
            for cookie in self._cookies.values():
                yield 'Set-Cookie', str(cookie)

    def coalesced_items(self) -> Generator[Tuple[str, str], None, None]:
        """Coalescence of headers does NOT garanty order of headers.
        It garanties the order of the header values, though.
        """
        if self._cookies:
            cookies = (str(cookie) for cookie in self._cookies.values())
        else:
            cookies = None

        keys = frozenset(self.keys())
        for header in keys:
            values = self.getall(header)
            if header == 'Set-Cookie' and cookies:
                values = [*values, *cookies]
            yield header, ', '.join(values)
        if 'Set-Cookie' not in self and cookies:
            yield 'Set-Cookie', ', '.join(cookies)


Finisher = Callable[[], None]


class Response(WSGICallable):

    __slots__ = ('status', 'body', 'headers', '_finishers')

    status: HTTPStatus
    body: Iterable
    headers: Headers
    _finishers: Optional[Deque[Finisher]]

    def __init__(self, status: HTTPCode = 200,
                 body: Optional[Iterable] = None,
                 headers: Optional[Headers] = None):
        self.status = HTTPStatus(status)
        self.body = body
        self.headers = Headers(headers or [])
        self._finishers = None

    @property
    def cookies(self):
        return self.headers.cookies

    def close(self):
        """Exhaust the list of finishers. No error is handled here.
        An exception will cause the closing operation to fail during
        the finishers iteration.
        """
        if self._finishers:
            while self._finishers:
                finisher = self._finishers.popleft()
                finisher()

    def add_finisher(self, task: Finisher):
        if self._finishers is None:
            self._finishers = Deque()
        self._finishers.append(task)

    def __iter__(self) -> Generator[bytes, None, None]:
        if self.status not in BODYLESS:
            if self.body is None:
                yield self.status.description.encode()
            elif isinstance(self.body, bytes):
                yield self.body
            elif isinstance(self.body, str):
                yield self.body.encode()
            elif isinstance(self.body, (Generator, Iterable)):
                yield from self.body
            else:
                raise TypeError(
                    f'Body of type {type(self.body)!r} is not supported.'
                )

    def __call__(self, environ: Environ,
                 start_response: StartResponse) -> Iterable:
        status = f'{self.status.value} {self.status.phrase}'
        start_response(status, list(self.headers.items()))
        return self

    @classmethod
    def redirect(cls, location, code: HTTPCode = 303,
                 body: Optional[Iterable] = None,
                 headers: Optional[Headers] = None):
        if code not in REDIRECT:
            raise ValueError(f"{code}: unknown redirection code.")
        if not headers:
            headers = {'Location': location}
        else:
            headers['Location'] = location
        return cls(code, body, headers)

    @classmethod
    def from_file_iterator(cls, filename: str, body: Iterable[bytes],
                           headers: Optional[Headers] = None):
        if headers is None:
            headers = {
                "Content-Disposition": f"attachment;filename={filename}"}
        elif "Content-Disposition" not in headers:
            headers["Content-Disposition"] = (
                f"attachment;filename={filename}")
        return cls(200, body, headers)

    @classmethod
    def to_json(cls, code: HTTPCode = 200, body: Optional[Any] = None,
                headers: Optional[Headers] = None):
        data = orjson.dumps(body)
        if headers is None:
            headers = {'Content-Type': 'application/json'}
        else:
            headers['Content-Type'] = 'application/json'
        return cls(code, data, headers)

    @classmethod
    def from_json(cls, code: HTTPCode = 200, body: str = '',
                  headers: Optional[Headers] = None):
        if headers is None:
            headers = {'Content-Type': 'application/json'}
        else:
            headers['Content-Type'] = 'application/json'
        return cls(code, body, headers)

    @classmethod
    def html(cls, code: HTTPCode = 200, body: str = '',
             headers: Optional[Headers] = None):
        if headers is None:
            headers = {'Content-Type': 'text/html; charset=utf-8'}
        else:
            headers['Content-Type'] = 'text/html; charset=utf-8'
        return cls(code, body, headers)



import pytest
from http import HTTPStatus
from horseman.http import HTTPError


def test_exception():
    with pytest.raises(ValueError) as exc:
        HTTPError('abc')
    assert str(exc.value) == "'abc' is not a valid HTTPStatus"

    exc = HTTPError(400)
    assert exc.status == HTTPStatus(400)
    assert exc.body == 'Bad request syntax or unsupported method'

    exc = HTTPError(400, body='')
    assert exc.status == HTTPStatus(400)
    assert exc.body == ''
    assert bytes(exc) == (
        b'HTTP/1.1 400 Bad Request\r\n'
        b'Content-Length: 0\r\n\r\n'
    )

    exc = HTTPError(404, body='I did not find anything')
    assert exc.status == HTTPStatus(404)
    assert exc.body == 'I did not find anything'
    assert bytes(exc) == (
        b'HTTP/1.1 404 Not Found\r\n'
        b'Content-Length: 23\r\n\r\nI did not find anything'
    )

    exc = HTTPError(404, body=b'Works with bytes body')
    assert exc.status == HTTPStatus(404)
    assert exc.body == 'Works with bytes body'
    assert bytes(exc) == (
        b'HTTP/1.1 404 Not Found\r\n'
        b'Content-Length: 21\r\n\r\nWorks with bytes body'
    )

    with pytest.raises(ValueError) as exc:
        HTTPError(200, 200)
    assert str(exc.value) == "Body must be string or bytes."



import pytest
import webtest
from unittest.mock import Mock
from http import HTTPStatus
from horseman.response import Response, file_iterator


def test_file_iterator(tmpdir):

    fpath = tmpdir / 'test.txt'
    with fpath.open('w+') as fd:
        fd.write('This is a sentence')

    fiter = file_iterator(fpath, chunk=3)
    chunks = list(fiter)
    assert chunks == [b'Thi', b's i', b's a', b' se', b'nte', b'nce']

    fiter = file_iterator(fpath)
    chunks = list(fiter)
    assert chunks == [b'This is a sentence']

    response = Response.from_file_iterator(
        'test.txt', file_iterator(fpath))
    assert response.status == 200
    assert list(response.headers.items()) == [
        ('Content-Disposition', 'attachment;filename=test.txt')
    ]
    assert list(response) == [b'This is a sentence']

    response = Response.from_file_iterator(
        'test.txt', file_iterator(fpath),
        headers={"Content-Type": "foo"}
    )
    assert list(response.headers.items()) == [
        ('Content-Type', 'foo'),
        ('Content-Disposition', 'attachment;filename=test.txt')
    ]

    response = Response.from_file_iterator(
        'test.txt', file_iterator(fpath),
        headers={"Content-Disposition": "foobar"}
    )
    assert list(response.headers.items()) == [
        ('Content-Disposition', 'foobar')
    ]


def test_can_set_status_from_numeric_value():
    response = Response(202)
    assert response.status == HTTPStatus.ACCEPTED


def test_raises_if_code_is_unknown():
    with pytest.raises(ValueError):
        Response(999)


def test_wrong_body_type():
    environ = {
        'PATH_INFO': '/'
    }
    response = Response(200, body=object())
    start_response = Mock()
    with pytest.raises(TypeError):
        list(response(environ, start_response))

    start_response.assert_called_with('200 OK', [])


def test_bytes_representation_bodyless():
    app = webtest.TestApp(
        Response(HTTPStatus.ACCEPTED)
    )
    response = app.get('/')
    assert response.status_int == 202
    assert response.body == (
        b'Request accepted, processing continues off-line'
    )
    assert list(response.headers.items()) == [('Content-Length', '47')]


def test_representation_with_body():
    app = webtest.TestApp(
        Response(HTTPStatus.OK, body="Super")
    )
    response = app.get('/')
    assert response.status_int == 200
    assert response.body == b'Super'
    assert list(response.headers.items()) == [('Content-Length', '5')]


def test_representation_bodyless_with_body():
    app = webtest.TestApp(
        Response(
            HTTPStatus.NO_CONTENT, body="Super")
    )
    response = app.get('/')
    assert response.status_int == 204
    assert response.body == b''
    assert list(response.headers.items()) == []


def test_304_no_content_type():
    app = webtest.TestApp(
        Response(HTTPStatus.NOT_MODIFIED)
    )
    response = app.get('/')
    assert response.status_int == 304
    assert response.body == b''
    assert list(response.headers.items()) == []


def test_1XX_no_content_type():
    app = webtest.TestApp(
        Response(HTTPStatus.CONTINUE)
    )
    response = app.get('/', status=100)
    assert response.status_int == 100
    assert response.body == b''
    assert list(response.headers.items()) == []


def test_json_response_headers():
    response = Response.from_json(body="{}")
    assert list(response.headers.items()) == [
        ('Content-Type', 'application/json')
    ]

    response = Response.from_json(
        body="{}", headers={"Content-Type": "text/html"})
    assert list(response.headers.items()) == [
        ('Content-Type', 'application/json')
    ]


def test_html_response():

    response = Response.html(body="<html></html>")
    assert list(response.headers.items()) == [
        ('Content-Type', 'text/html; charset=utf-8')
    ]

    response = Response.html(
        body="{}", headers={"Content-Type": "text/plain"})
    assert list(response.headers.items()) == [
        ('Content-Type', 'text/html; charset=utf-8')
    ]


def test_json_response():
    structure = {
        'Horseman': 'headless',
        'python3.8': True,
        'version': 0.1
    }

    app = webtest.TestApp(
        Response.to_json(body=structure)
    )
    response = app.get('/')
    assert response.status_int == 200
    assert response.body == (
        b'{"Horseman":"headless","python3.8":true,"version":0.1}')
    assert list(response.headers.items()) == [
        ('Content-Type', 'application/json'),
        ('Content-Length', '54')
    ]

    app = webtest.TestApp(
        Response.to_json(
            body=structure, headers={'Custom-Header': 'Test'}))
    response = app.get('/')
    assert response.status_int == 200
    assert response.body == (
        b'{"Horseman":"headless","python3.8":true,"version":0.1}')
    assert list(response.headers.items()) == [
        ('Custom-Header', 'Test'),
        ('Content-Type', 'application/json'),
        ('Content-Length', '54')
    ]

    app = webtest.TestApp(
        Response.to_json(
            HTTPStatus.ACCEPTED, body=structure,
            headers={'Custom-Header': 'Test'})
    )
    response = app.get('/')
    assert response.status_int == 202
    assert response.body == (
        b'{"Horseman":"headless","python3.8":true,"version":0.1}')
    assert list(response.headers.items()) == [
        ('Custom-Header', 'Test'),
        ('Content-Type', 'application/json'),
        ('Content-Length', '54')
    ]

    app = webtest.TestApp(
        Response.to_json(
            HTTPStatus.ACCEPTED, body=structure,
            headers={'Content-Type': 'wrong/content'})
    )
    response = app.get('/')
    assert response.status_int == 202
    assert response.body == (
        b'{"Horseman":"headless","python3.8":true,"version":0.1}')
    assert list(response.headers.items()) == [
        ('Content-Type', 'application/json'),
        ('Content-Length', '54')
    ]

    app = webtest.TestApp(
        Response.to_json(
            HTTPStatus.ACCEPTED, body=structure,
            headers={})
    )
    response = app.get('/')
    assert response.status_int == 202
    assert response.body == (
        b'{"Horseman":"headless","python3.8":true,"version":0.1}')
    assert list(response.headers.items()) == [
        ('Content-Type', 'application/json'),
        ('Content-Length', '54')
    ]


def test_json_errors():
    with pytest.raises(TypeError):
        Response.to_json(body=object())


def test_redirect():
    response = Response.redirect('/test')
    assert list(response.headers.items()) == [('Location', '/test')]
    assert response.body is None
    assert response.status == 303
    assert webtest.TestApp(response).get('/').body == (
        b'Object moved -- see Method and URL list')

    response = Response.redirect('/test', code=301)
    assert list(response.headers.items()) == [('Location', '/test')]
    assert response.body is None
    assert response.status == 301
    assert webtest.TestApp(response).get('/').body == (
        b'Object moved permanently -- see URI list'
    )

    response = Response.redirect(
        '/test', code=301, headers={"Location": "/outside"})
    assert list(response.headers.items()) == [('Location', '/test')]
    assert response.body is None
    assert response.status == 301
    assert webtest.TestApp(response).get('/').body == (
        b'Object moved permanently -- see URI list'
    )


def test_invalid_redirect():
    with pytest.raises(ValueError) as exc:
        Response.redirect('/test', code=400)
    assert str(exc.value) == '400: unknown redirection code.'


def test_response_cookies():
    response = Response()
    assert response.cookies == {}
    assert response.cookies is response.headers.cookies


def test_finishers():
    from collections import deque

    response = Response()
    assert response._finishers is None
    assert response.close() is None

    task = Mock()
    response.add_finisher(task)
    assert response._finishers is not None
    assert isinstance(response._finishers, deque)
    assert len(response._finishers) == 1
    assert task.called is False

    response.close()
    assert task.called is True
    assert task.assert_called_with() is None


def test_finishers_order():
    calls = []

    def tasker(num):
        def task():
            calls.append(num)
        return task

    response = Response()
    response.add_finisher(tasker(1))
    response.add_finisher(tasker(2))
    response.add_finisher(tasker(3))
    response.close()

    assert calls == [1, 2, 3]

    response.add_finisher(tasker(3))
    response.add_finisher(tasker(2))
    response.add_finisher(tasker(1))
    response.close()
    assert calls == [1, 2, 3, 3, 2, 1]


def test_exception_finisher():

    calls = []

    def tasker(num):
        def task():
            if num == 2:
                raise RuntimeError('I do not want a 2')
            calls.append(num)
        return task

    response = Response()
    response.add_finisher(tasker(1))
    response.add_finisher(tasker(2))
    response.add_finisher(tasker(3))

    with pytest.raises(RuntimeError):
        response.close()

    assert calls == [1]
    assert len(response._finishers) == 1  # tasker(3) never popped.

    # We can resume by calling close again.
    # Note that the task responsible for the exception is lost at
    # thus point.
    response.close()
    assert calls == [1, 3]
    assert len(response._finishers) == 0



import pytest
from io import BytesIO
from webtest.app import TestApp as App
from horseman.parsers import parser
from horseman.http import HTTPError


BAD_MULTIPART = (
    b"""--foo\r\nContent->Disposition: form-data; """
    b"""name="text1"\r\n\r\n'abc\r\n--foo--\r\n"""
)

BAD_MULTIPART_NO_CONTENT_DISPOSITION = (
    b"""------------a_BoUnDaRy7283067873172754$\r\n"""
    b"""Content-Disposition: ; name="test"\r\n\r\nsome value\r\n"""
    b"""------------a_BoUnDaRy7283067873172754$--\r\n"""
)


def test_multipart():

    app = App(None)
    content_type, body = app.encode_multipart([
        ('test', 'some value'),
        ('test', 'some other value'),
        ('foo', 'bar')
    ], [])

    data = parser(BytesIO(body), content_type)
    assert len(data.form) == 2
    assert list(data.form.items()) == [
        ('test', ['some value', 'some other value']),
        ('foo', ['bar'])
    ]
    assert data.form.get('foo') == 'bar'
    assert data.form.getlist('test') == ['some value', 'some other value']


def test_multipart_empty_files_empty_name():
    app = App(None)
    content_type, body = app.encode_multipart(
        [],
        [('test', "", b'', "application/octet")],
    )
    data = parser(BytesIO(body), content_type)
    assert not data.files

    content_type, body = app.encode_multipart(
        [],
        [('test', "", b'', "application/octet"),
         ('test2', "", b'', "application/octet")],
    )
    data = parser(BytesIO(body), content_type)
    assert not data.files


def test_multipart_empty_files_mixed():
    app = App(None)
    content_type, body = app.encode_multipart(
        [],
        [('test', "name", b'', "application/octet"),
         ('test2', "", b'content', "application/octet"),
         ('test3', "", b'', "application/octet")],
    )
    data = parser(BytesIO(body), content_type)
    assert 'test3' not in data.files
    assert data.files['test'][0].filename == "name"
    assert data.files['test2'][0].read() == b'content'
    assert data.files['test2'][0].filename != ''


def test_multipart_empty_filename_generation():
    app = App(None)
    content_type, body = app.encode_multipart(
        [],
        [('test', "", b'content', "application/octet"),
         ('test', "", b'content', "application/octet")]
    )
    data = parser(BytesIO(body), content_type)
    assert data.files['test'][0].filename != data.files['test'][1].filename
    assert data.files['test'][0] != ''


def test_multipart_files():

    app = App(None)
    content_type, body = app.encode_multipart(
        [('test', b'some value')],
        [('files', "baz-\xe9.png", b'abcdef', 'image/png'),
         ('files', "MyText.txt", b'ghi', 'text/plain')]
    )

    data = parser(BytesIO(body), content_type)
    uploaded = data.files.getlist('files')
    assert len(uploaded) == 2
    assert uploaded[0].filename == 'baz-é.png'
    assert uploaded[0].content_type == b'image/png'
    assert uploaded[0].read() == b'abcdef'
    assert uploaded[1].filename == 'MyText.txt'
    assert uploaded[1].content_type == b'text/plain'
    assert uploaded[1].read() == b'ghi'


def test_wrong_multipart():
    with pytest.raises(HTTPError) as exc:
        parser(BytesIO(b'test'), "multipart/form-data")
    assert exc.value.status == 400
    assert exc.value.body == 'Missing boundary in Content-Type.'

    with pytest.raises(HTTPError) as exc:
        parser(BytesIO(BAD_MULTIPART),
               "multipart/form-data; boundary=--foo")
    assert exc.value.status == 400
    assert exc.value.body == 'Unparsable multipart body.'


def test_wrong_multipart_no_content_disposition():

    with pytest.raises(HTTPError) as exc:
        parser(BytesIO(BAD_MULTIPART_NO_CONTENT_DISPOSITION),
               'multipart/form-data; '
               'boundary=----------a_BoUnDaRy7283067873172754$')

    assert exc.value.status == 400
    assert exc.value.body == ('Unparsable multipart body.')



import pytest
from webtest.app import TestRequest as Request
from horseman.http import Query


def test_query():
    q = Query.from_environ({})
    assert isinstance(q, Query)
    assert len(q) == 0


def test_float_should_cast_to_float():
    request = Request.blank('/?key=1', method='GET')
    query = Query.from_environ(request.environ)
    assert query.float('key') == 1.0

    request = Request.blank('/?key=42.1', method='GET')
    query = Query.from_environ(request.environ)
    assert query.float('key') == 42.1

    request = Request.blank('/?key=-75', method='GET')
    query = Query.from_environ(request.environ)
    assert query.float('key') == -75

    request = Request.blank('/?key=-75.5', method='GET')
    query = Query.from_environ(request.environ)
    assert query.float('key') == -75.5

    request = Request.blank('/?key=-75,5', method='GET')
    query = Query.from_environ(request.environ)
    with pytest.raises(ValueError):
        query.int('key')

    request = Request.blank('/?key=dang!', method='GET')
    query = Query.from_environ(request.environ)
    with pytest.raises(ValueError):
        query.int('key')


def test_bool_should_cast_to_boolean():
    for value in Query.TRUE_STRINGS:
        request = Request.blank(f'/?key={value}', method='GET')
        query = Query.from_environ(request.environ)
        assert query.bool('key') is True

    for value in Query.FALSE_STRINGS:
        request = Request.blank(f'/?key={value}', method='GET')
        query = Query.from_environ(request.environ)
        assert query.bool('key') is False

    for value in Query.NONE_STRINGS:
        request = Request.blank(f'/?key={value}', method='GET')
        query = Query.from_environ(request.environ)
        assert query.bool('key') is None

    request = Request.blank('/?key=Z', method='GET')
    query = Query.from_environ(request.environ)
    with pytest.raises(ValueError) as exc:
        query.bool('key')
    assert str(exc.value) == "Can't cast 'z' to boolean."

    q = Query({'foo': True, 'bar': False, 'crux': None})
    assert q.bool('foo') is True
    assert q.bool('bar') is False
    assert q.bool('crux') is None


def test_int_should_cast_to_int():
    request = Request.blank('/?key=1', method='GET')
    query = Query.from_environ(request.environ)
    assert query.int('key') == 1

    request = Request.blank('/?key=42', method='GET')
    query = Query.from_environ(request.environ)
    assert query.int('key') == 42

    request = Request.blank('/?key=-75', method='GET')
    query = Query.from_environ(request.environ)
    assert query.int('key') == -75

    request = Request.blank('/?key=75.5', method='GET')
    query = Query.from_environ(request.environ)
    with pytest.raises(ValueError):
        query.int('key')

    request = Request.blank('/?key=dang!', method='GET')
    query = Query.from_environ(request.environ)
    with pytest.raises(ValueError):
        query.int('key')



import pytest
from io import BytesIO
from horseman.parsers import Data, json_parser


def test_empty_json():
    body = BytesIO(b'')
    with pytest.raises(ValueError) as exc:
        json_parser(body, 'application/json')
    assert str(exc.value) == "The body of the request is empty."


def test_json():
    body = BytesIO(b'{"foo": "bar"}')
    data = json_parser(body, 'application/json')
    assert data == Data(
        form=None,
        files=None,
        json={'foo': 'bar'}
    )


def test_json_charset():
    body = BytesIO('{"name": "Älfùr"}'.encode('utf-8'))
    data = json_parser(body, 'application/json')
    assert data == Data(
        form=None,
        files=None,
        json={'name': 'Älfùr'}
    )

    body = BytesIO('{"name": "Älfùr"}'.encode('latin-1'))
    with pytest.raises(ValueError) as exc:
        json_parser(body, 'application/json')
    assert str(exc.value) == (
        "'utf-8' codec can't decode byte 0xc4 in position 10: "
        "invalid continuation byte")

    body = BytesIO('{"name": "Älfùr"}'.encode('latin-1'))
    data = json_parser(body, 'application/json', charset='latin-1')
    assert data == Data(
        form=None,
        files=None,
        json={'name': 'Älfùr'}
    )


def test_wrong_json():
    body = BytesIO(b'foo')
    with pytest.raises(ValueError) as exc:
        json_parser(body, 'application/json')
    assert str(exc.value) == "Unparsable JSON body."



import sys
import re
from abc import ABC, abstractmethod
from typing import TypeVar
from horseman.response import Response
from horseman.http import HTTPError
from horseman.types import (
    WSGICallable, Environ, StartResponse, ExceptionInfo)


Data = TypeVar('Data')
slashes_normalization = re.compile(r"/+")


class Overhead(ABC):
    """WSGI Environ Overhead aka Request representation
    This object contains everything needed to handle a request.
    It can carry DB connectors, parsed data and other utils.
    """
    data: Data
    environ: Environ

    @abstractmethod
    def extract(self) -> Data:
        """Extracts the data from the incoming HTTP request.
        """


class APIView:
    """View with methods to act as HTTP METHOD dispatcher.
    Method names of the class must be a valid uppercase HTTP METHOD name.
    example : OPTIONS, GET, POST
    """

    def __call__(self, overhead: Overhead) -> Response:
        method = overhead.environ['REQUEST_METHOD'].upper()
        if worker := getattr(self, method, None):
            return worker(overhead)

        # Method not allowed
        return Response(405)


class Node(WSGICallable):

    @abstractmethod
    def resolve(self, path_info: str, environ: Environ) -> WSGICallable:
        """Resolves the path into a wsgi callable (eg. Response).
        If nothing was found, returns None or a WSGI callable corresponding
        to the HTTP Error (404, 405, 406).
        """

    def __call__(self, environ: Environ, start_response: StartResponse):
        # according to PEP 3333 the native string representing PATH_INFO
        # (and others) can only contain unicode codepoints from 0 to 255,
        # which is why we need to decode to latin-1 instead of utf-8 here.
        # We transform it back to UTF-8
        # Note that it's valid for WSGI server to omit the value if it's
        # empty.
        path_info = environ.get(
            'PATH_INFO', '').encode('latin-1').decode('utf-8')
        if path_info:
            # Normalize the slashes to avoid things like '//test'
            path_info = slashes_normalization.sub("/", path_info)
        try:
            response = self.resolve(path_info, environ)
            if response is None:
                response = Response(404)

        except HTTPError as error:
            # FIXME: Log.
            response = Response(error.status, error.body)
        return response(environ, start_response)


class SentryNode(Node):

    @abstractmethod
    def handle_exception(self, exc_info: ExceptionInfo, environ: Environ):
        """This method handles exceptions happening while the
        application is trying to render/process/interpret the request.
        """

    def __call__(self, environ: Environ, start_response: StartResponse):
        iterable = None
        try:
            iterable = super().__call__(environ, start_response)
            yield from iterable
        except Exception:
            exc_info = sys.exc_info()
            self.handle_exception(exc_info, environ)
            exc_info = None
            raise
        finally:
            if hasattr(iterable, 'close'):
                try:
                    iterable.close()
                except Exception:
                    exc_info = sys.exc_info()
                    self.handle_exception(exc_info, environ)
                    exc_info = None
                    raise



from horseman.http import Cookies


def test_request_parse_cookies():
    from webtest.app import TestRequest as Request

    # A simple cookie
    request = Request(environ={}, cookies={'key': 'value'})
    cookies = Cookies.from_environ(request.environ)
    assert cookies['key'] == 'value'

    # No cookie
    request = Request(environ={})
    cookies = Cookies.from_environ(request.environ)
    assert not cookies



from frozendict import frozendict
from typing import (
    cast, Optional, Union, Any, Mapping, List, Iterable, Tuple, Dict)


Pairs = Iterable[Tuple[str, Any]]


class FormData(Dict[str, List[Any]]):

    def __init__(self, data: Optional[Union['FormData', Dict, Pairs]] = None):
        if data is not None:
            if isinstance(data, FormData):
                super().__init__(data)
            elif isinstance(data, Mapping):
                for key, value in data.items():
                    self.add(key, value)
            else:
                value = cast(Pairs, data)
                for key, value in value:
                    self.add(key, value)

    def get(self, name: str, default: Optional[Any] = None) -> Any:
        """Return the first value of the found list.
        """
        return super().get(name, [default])[0]

    def getlist(self, name: str, default: Optional[Any] = None) -> List[Any]:
        """Return the value list
        """
        return super().get(name, default)

    def pairs(self) -> Pairs:
        for key, values in self.items():
            for value in values:
                yield key, value

    def add(self, name: str, value: Any) -> None:
        if name in self:
            self[name].append(value)
        else:
            self[name] = [value]

    def to_dict(self, frozen=True):
        impl = frozendict if frozen else dict
        return impl(
            {k: (v[0] if len(v) == 1 else v) for k, v in self.items()}
        )


class TypeCastingDict(FormData):
    TRUE_STRINGS = {'t', 'true', 'yes', '1', 'on'}
    FALSE_STRINGS = {'f', 'false', 'no', '0', 'off'}
    NONE_STRINGS = {'n', 'none', 'null'}

    def bool(self, key: str, default=...):
        value = self.get(key, default)
        if value in (True, False, None):
            return value
        value = value.lower()
        if value in self.TRUE_STRINGS:
            return True
        elif value in self.FALSE_STRINGS:
            return False
        elif value in self.NONE_STRINGS:
            return None
        raise ValueError(f"Can't cast {value!r} to boolean.")

    def int(self, key: str, default=...):
        return int(self.get(key, default))

    def float(self, key: str, default=...):
        return float(self.get(key, default))



import re
from typing import Mapping
from collections import UserDict
from horseman.meta import Node
from horseman.types import WSGICallable, Environ


class Mapping(Node, UserDict, Mapping[str, WSGICallable]):

    NORMALIZE = re.compile('//+')

    @classmethod
    def normalize(cls, path: str):
        if not isinstance(path, str):
            raise ValueError(f'{cls} accepts only str keys.')
        if not path.startswith('/'):
            raise ValueError(f"Path must start with '/', got {path!r}")
        return cls.NORMALIZE.sub('/', path)

    def __setitem__(self, path: str, script: WSGICallable):
        super().__setitem__(self.normalize(path), script)

    def resolve(self, path_info: str, environ: Environ) -> WSGICallable:
        for script_name in sorted(self.keys(), key=len, reverse=True):
            if path_info.startswith(script_name):
                script = self[script_name]
                name = script_name.rstrip('/')
                environ['SCRIPT_NAME'] += name
                environ['PATH_INFO'] = path_info[len(name):]
                return script



import pytest
from io import BytesIO
from horseman.parsers import BodyParser, Data
from horseman.http import HTTPError, ContentType


def test_parser_invalid_registration():
    parser = BodyParser()

    with pytest.raises(ValueError) as exc:

        @parser.register('foo')
        def test(body, mimetype, **options):
            pass

    assert str(exc.value) == "'foo' is not a valid MIME Type."


def test_parser_registration():
    parser = BodyParser()

    @parser.register('foo/bar')
    def test(body, mimetype, **options):
        pass

    assert parser.get('foo/bar') == test


def test_parser_registration_override():
    parser = BodyParser()

    @parser.register('foo/bar')
    def test(body, mimetype, **options):
        pass

    assert parser.get('foo/bar') == test

    @parser.register('foo/bar')
    def test2(body, mimetype, **options):
        pass

    assert parser.get('foo/bar') == test2


def test_unknown_parser():
    parser = BodyParser()

    with pytest.raises(HTTPError) as exc:
        parser(BytesIO(b'body'), 'foo/bar')

    assert exc.value.status == 400
    assert exc.value.body == "Unknown content type: 'foo/bar'."

    @parser.register('foo/bar')
    def test(body, mimetype, **options):
        return Data()

    data = parser(BytesIO(b'body'), 'foo/bar')
    assert isinstance(data, Data)

    data = parser(BytesIO(b'body'), 'foo/bar; charset=UTF-8')
    assert isinstance(data, Data)


def test_parser_with_mimetype():
    parser = BodyParser()
    contenttype = ContentType('foo/bar', {})

    @parser.register('foo/bar')
    def test(body, mimetype, **options):
        return Data()

    data = parser(BytesIO(b'body'), contenttype)
    assert isinstance(data, Data)



import pytest
import hamcrest
from frozendict import frozendict
from horseman.datastructures import FormData


def test_data_wrong_values():

    with pytest.raises(ValueError):
        FormData('test')

    with pytest.raises(ValueError):
        FormData({"test"})

    with pytest.raises(ValueError):
        FormData([('a', 'b', 'c')])


def test_data_init():
    # no data
    fd = FormData()
    assert fd == {}

    # List of tuples
    fd = FormData([('a', 1), ('b', 7), ('a', -3)])
    assert fd == {'a': [1, -3], 'b': [7]}

    # tuple of tuples
    fd = FormData((('a', 1), ('b', 7), ('a', -3)))
    assert fd == {'a': [1, -3], 'b': [7]}

    # set of tuples: no garanteed orders
    fd = FormData({('a', 1), ('b', 7), ('a', -3)})
    hamcrest.assert_that(fd, hamcrest.has_entries({
        'a': hamcrest.contains_inanyorder(-3, 1),
        'b': [7]
    }))

    # dict
    fd = FormData({'a': 1, 'b': 7})
    assert fd == {'a': [1], 'b': [7]}

    # formdata
    fd = FormData([('a', 1), ('b', 7), ('a', -3)])
    fd2 = FormData(fd)
    assert fd is not fd2
    assert fd == fd2
    assert fd2 == {'a': [1, -3], 'b': [7]}


def test_access():
    fd = FormData([('a', 1), ('b', 7), ('a', -3)])
    assert fd.get('a') == 1
    assert fd.get('b') == 7
    assert fd.getlist('a') == [1, -3]
    assert fd.getlist('b') == [7]
    assert list(fd.pairs()) == [
        ('a', 1), ('a', -3), ('b', 7)
    ]


def test_casting():
    fd = FormData([('a', 1), ('b', 7), ('a', -3)])
    assert fd == {'a': [1, -3], 'b': [7]}

    dictified = fd.to_dict()
    hamcrest.assert_that(dictified, hamcrest.has_entries({
        'a': hamcrest.contains_inanyorder(-3, 1),
        'b': 7
    }))
    assert isinstance(dictified, frozendict)

    dictified = fd.to_dict(frozen=False)
    hamcrest.assert_that(dictified, hamcrest.has_entries({
        'a': hamcrest.contains_inanyorder(-3, 1),
        'b': 7
    }))
    assert isinstance(dictified, dict)


def test_empty_value():
    fd = FormData([
        ('email', 'ck@novareto.de'), ('contact', '')])
    assert fd == {'contact': [''], 'email': ['ck@novareto.de']}
    assert fd.to_dict() == {
        'contact': '',
        'email': 'ck@novareto.de'
    }



from http import HTTPStatus
from types import TracebackType
from typing import (
    Any,
    ByteString,
    Callable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Literal,
)

HTTPMethod = Literal[
    "GET", "HEAD", "PUT", "DELETE", "PATCH", "POST", "OPTIONS"
]

Charset = TypeVar('Charset', str, bytes)
MIMEType = TypeVar('MIMEType', str, bytes)
HTTPCode = TypeVar('HTTPCode', HTTPStatus, int)
StatusCode = TypeVar('StatusCode', str, bytes)
URLParameter = TypeVar('URLParameter')

Environ = Mapping[str, Any]
ExceptionInfo = Tuple[Exception, Any, TracebackType]
ResponseHeaders = Sequence[Tuple[str, str]]
StartResponse = Callable[
    [StatusCode, ResponseHeaders, Optional[ExceptionInfo]],
    Optional[Callable[[ByteString], None]]
]
WSGICallable = Callable[[Environ, StartResponse], Iterable[bytes]]



import hamcrest
from horseman.response import Headers


def test_empty_headers():
    headers = Headers()
    assert list(headers.items()) == []
    assert len(headers) == 0


def test_add_headers():
    headers = Headers({'Link': 'test'})
    assert list(headers.items()) == [('Link', 'test')]
    assert len(headers) == 1

    headers.add('Link', 'Another test')
    assert list(headers.items()) == [
        ('Link', 'test'),
        ('Link', 'Another test')
    ]
    assert len(headers) == 2


def test_cookies():
    headers = Headers()
    headers.cookies.set('test', "{'this': 'is json'}")
    assert 'test' in headers.cookies
    assert list(headers.items()) == [
        ('Set-Cookie', 'test="{\'this\': \'is json\'}"; Path=/')
    ]


def test_headers_coalescence():
    """Coalescence of headers does NOT garanty order of headers.
    It garanties the order of the header values, though.
    """
    headers = Headers([
        ('Link', 'test'),
        ('X-Robots-Tag', 'noarchive'),
        ('X-Robots-Tag', 'google: noindex, nosnippet')
    ])

    hamcrest.assert_that(
        list(headers.coalesced_items()),
        hamcrest.contains_inanyorder(
            ('Link', 'test'),
            ('X-Robots-Tag', 'noarchive, google: noindex, nosnippet'),
        )
    )

    headers.add('X-Robots-Tag', 'otherbot: noindex')
    hamcrest.assert_that(
        list(headers.coalesced_items()),
        hamcrest.contains_inanyorder(
            ('Link', 'test'),
            ('X-Robots-Tag', 'noarchive, google: noindex, nosnippet, '
                             'otherbot: noindex'),
        )
    )

    assert list(headers.items()) == [
        ('Link', 'test'),
        ('X-Robots-Tag', 'noarchive'),
        ('X-Robots-Tag', 'google: noindex, nosnippet'),
        ('X-Robots-Tag', 'otherbot: noindex'),
    ]


def test_headers_coalescence_with_cookies():
    headers = Headers()
    headers.cookies.set('test', "{'this': 'is json'}")
    headers.add('X-Robots-Tag', 'otherbot: noindex')

    hamcrest.assert_that(
        list(headers.coalesced_items()),
        hamcrest.contains_inanyorder(
            ('X-Robots-Tag', 'otherbot: noindex'),
            ('Set-Cookie', 'test="{\'this\': \'is json\'}"; Path=/')
        )
    )

    headers.add('Set-Cookie', 'other=foobar')
    hamcrest.assert_that(
        list(headers.coalesced_items()),
        hamcrest.contains_inanyorder(
            ('X-Robots-Tag', 'otherbot: noindex'),
            ('Set-Cookie', 'other=foobar, '
                           'test="{\'this\': \'is json\'}"; Path=/')
        )
    )



from io import BytesIO
from multifruits import Parser, extract_filename, parse_content_disposition
from horseman.datastructures import FormData


class Multipart:
    """Responsible of the parsing of multipart encoded body.
    """

    __slots__ = (
        'form',
        'files',
        '_parser',
        '_current',
        '_current_headers',
        '_current_params'
    )

    def __init__(self, content_type: str):
        self._parser = Parser(self, content_type.encode())
        self.form = FormData()
        self.files = FormData()

    def feed_data(self, data: bytes):
        self._parser.feed_data(data)

    def on_part_begin(self):
        self._current_headers = {}

    def on_header(self, field: bytes, value: bytes):
        self._current_headers[field] = value

    def on_headers_complete(self):
        disposition_type, params = parse_content_disposition(
            self._current_headers.get(b'Content-Disposition'))
        if not disposition_type:
            raise ValueError('Content-Disposition is missing.')

        self._current_params = params
        if b'Content-Type' in self._current_headers:
            self._current = BytesIO()
            self._current.filename = extract_filename(params)
            self._current.size = 0
            self._current.content_type = self._current_headers[b'Content-Type']
            self._current.params = params
        else:
            self._current = ''

    def on_data(self, data: bytes):
        if b'Content-Type' in self._current_headers:
            self._current.write(data)
            self._current.size += len(data)
        else:
            self._current += data.decode()

    def on_part_complete(self):
        name = self._current_params.get(b'name', b'').decode()
        if b'Content-Type' in self._current_headers:
            self._current.seek(0)
            if not self._current.filename:
                if not self._current.getbuffer().nbytes:
                    # This is an empty file with no name
                    # We do *not* save it.
                    self._current = None
                    return
                # at this point, we've got content but no name.
                # generate one.
                self._current.filename = str(id(self._current))
            self.files.add(name, self._current)
        else:
            self.form.add(name, self._current)
        self._current = None



import pytest
from horseman.meta import Node
from horseman.mapping import Mapping
from horseman.response import Response


def basic_app(environ, start_fn):
    start_fn('200 OK', [('Content-Type', 'text/plain')])
    return [b"Hello World!\n"]


def other_app(environ, start_fn):
    start_fn('200 OK', [('Content-Type', 'text/plain')])
    return [b"Something else\n"]


def test_mapping():
    node = Mapping()
    assert isinstance(node, Node)


def test_normalizing():
    with pytest.raises(ValueError):
        assert Mapping.normalize(1)
    assert Mapping.normalize('/test') == '/test'
    assert Mapping.normalize("/") == '/'
    assert Mapping.normalize("////") == '/'
    assert Mapping.normalize("//") == '/'
    assert Mapping.normalize("//") == '/'
    assert Mapping.normalize('/test//') == '/test/'


def test_mapping_instanciation():
    with pytest.raises(ValueError) as exc:
        Mapping(test=1)
    assert str(exc.value) == "Path must start with '/', got 'test'"

    node = Mapping({"/test": basic_app})
    assert node == {'/test': basic_app}

    node = Mapping({"/": basic_app})
    assert node == {'/': basic_app}

    node = Mapping({"////": basic_app})
    assert node == {'/': basic_app}


def test_mapping_update():
    node = Mapping({"/test": basic_app})
    with pytest.raises(ValueError):
        node.update({1: basic_app})

    node.update({'/test': other_app})
    assert node['/test'] is other_app


def test_mapping_set_default():
    node = Mapping()
    node.setdefault('/test', basic_app)
    assert node['/test'] is basic_app


def test_mapping_items():
    node = Mapping({"/test": basic_app, '/other': other_app})
    assert list(node.items()) == [
        ('/test', basic_app),
        ('/other', other_app)
    ]


def test_mapping_repr():
    node = Mapping({"/test": basic_app, '/other': other_app})
    assert repr(node) == (
        f"{{'/test': {basic_app}, '/other': {other_app}}}"
    )


def test_mapping_set_del_item():
    node = Mapping()
    node['/'] = basic_app
    assert node == {'/': basic_app}

    with pytest.raises(ValueError) as exc:
        node['test'] = basic_app
    assert str(exc.value) == "Path must start with '/', got 'test'"

    del node['/']
    assert node == {}

    node['/'] = 1
    node['/'] = basic_app
    assert node == {'/': basic_app}

    script = node.pop('/')
    assert script is basic_app
    assert node == {}


def test_mapping_resolve():
    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/no'}
    node = Mapping({"/test": basic_app})
    assert node.resolve('/no', environ) is None
    assert environ == {'PATH_INFO': '/no', 'SCRIPT_NAME': ''}

    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/test'}
    node = Mapping({"/test": basic_app})
    assert node.resolve('/test', environ) is basic_app
    assert environ == {'PATH_INFO': '', 'SCRIPT_NAME': '/test'}

    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/'}
    node = Mapping({"/": basic_app})
    assert node.resolve('/', environ) is basic_app
    assert environ == {'PATH_INFO': '/', 'SCRIPT_NAME': ''}


def test_mapping_competing_apps():
    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/backend'}
    node = Mapping({"/": basic_app, "/backend": other_app})
    assert node.resolve('/backend', environ) is other_app
    assert environ == {'PATH_INFO': '', 'SCRIPT_NAME': '/backend'}


def test_nested_mapping():
    from unittest.mock import Mock

    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/no'}
    node = Mapping({"/some":  Mapping({'/thing': basic_app})})
    response = node(environ, Mock())
    assert isinstance(response, Response)
    assert response.status == 404
    assert environ == {'SCRIPT_NAME': '', 'PATH_INFO': '/no'}

    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/some'}
    node = Mapping({"/some":  Mapping({'/thing': basic_app})})
    response = node(environ, Mock())
    assert response.status == 404
    assert environ == {'SCRIPT_NAME': '/some', 'PATH_INFO': ''}

    environ = {'SCRIPT_NAME': '', 'PATH_INFO': '/some/thing'}
    node = Mapping({"/some":  Mapping({'/thing': basic_app})})
    response = node(environ, Mock())
    assert response == [b'Hello World!\n']
    assert environ == {'PATH_INFO': '', 'SCRIPT_NAME': '/some/thing'}



import cgi
from http import HTTPStatus
from typing import NamedTuple, Dict, AnyStr
from urllib.parse import parse_qsl
from biscuits import Cookie, parse
from horseman.datastructures import TypeCastingDict
from horseman.types import HTTPCode, MIMEType


class HTTPError(Exception):

    def __init__(self, status: HTTPCode, body: AnyStr = ...):
        self.status = HTTPStatus(status)
        body = self.status.description if body is ... else body
        if isinstance(body, bytes):
            body = body.decode('utf-8')
        elif not isinstance(body, str):
            raise ValueError('Body must be string or bytes.')
        self.body: str = body

    def __bytes__(self):
        return ('HTTP/1.1 {status} {phrase}\r\n'
                'Content-Length: {length}\r\n\r\n{body}').format(
                    status=self.status.value,
                    phrase=self.status.phrase,
                    length=len(self.body),
                    body=self.body
                ).encode()


class ContentType(NamedTuple):
    mimetype: MIMEType
    options: dict

    @classmethod
    def from_http_header(cls, header: str):
        return cls(*cgi.parse_header(header))


class Cookies(Dict[str, Cookie]):
    """A Cookies management class, built on top of biscuits."""

    def set(self, name, *args, **kwargs):
        self[name] = Cookie(name, *args, **kwargs)

    @staticmethod
    def from_environ(environ: dict):
        return parse(environ.get('HTTP_COOKIE', ''))


class Query(TypeCastingDict):

    @classmethod
    def from_value(cls, value: str):
        return cls(parse_qsl(
            value, keep_blank_values=True, strict_parsing=True))

    @classmethod
    def from_environ(cls, environ: dict):
        qs = environ.get('QUERY_STRING', '')
        if qs:
            return cls.from_value(qs)
        return cls()



from setuptools import setup


setup(
    name='horseman',
    install_requires=[
        'biscuits >= 0.3.0',
        'multidict >= 5.1',
        'multifruits >= 0.1.5',
        'orjson >= 3.5',
        'frozendict >= 2.0',
    ],
    extras_require={
        'test': [
            'WebTest',
            'pytest',
            'pyhamcrest'
        ]
    }
)



import logging
from bjoern import run
from horseman.meta import SentryNode, Overhead, APIView
from horseman.response import Response


class Request(Overhead):

    data = None

    def __init__(self, environ):
        self.environ = environ

    def extract(self):
        self.data = 'somedata'


class View(APIView):

    def GET(self, overhead):
        return Response.to_json(200, {"Result": "OK"})


VIEWS = {
    "/": View()
}


class RootNode(SentryNode):

    def resolve(self, path_info, environ):
        if view := VIEWS.get(path_info):
            request = Request(environ)
            return view(request)

    def handle_exception(self, exc_info, environ):
        logging.error(exc_info)


run(
    host="0.0.0.0",
    port=8080,
    reuse_port=True,
    wsgi_app=RootNode(),
)
